##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
  "token": {
    "token": "240a3408",
    "expired_time": "2014-09-03T17:15:45+0800",
    "max_idle_interval": "1800000"
  }
}

ONE = ALL



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import os
from importlib import import_module

_PATH = os.path.abspath(os.path.dirname(__file__))
RESOURCES = \
    [resource for resource in os.listdir(_PATH) if os.path.isdir(
        os.path.join(_PATH, resource)
        )
     ]
RESOURCE_NAME_CLASS_MAP = {}

for re in RESOURCES:
    if re != '__pycache__':
        RESOURCE_NAME_CLASS_MAP.update(
            import_module('{0}.{1}'.format(__name__,
                                           re)).RESOURCE_NAME_CLASS_MAP
        )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
       "server":
       {
           "status": "ok",
           "code": "CMUC00183I",
           "message": "Operation done successfully."
       },
       "data":
       {
           "eserep":
           [
               {
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/pools/P1/eserep"
                   },
                   "cap": "136286",
                   "capalloc": "0",
                   "capavail": "136286",
                   "overprovisioned": "0.0",
                   "repcapthreshold": "0",
                   "pool":
                   {
                       "id": "P1",
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/pools/P1"
                       }
                   }
               }
           ]
       }
    }

ONE = ALL



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
IO Enclosure interface.
"""

from .common.types import DS8K_IOENCLOSURE
from .common.base import Base, ReadOnlyManager


class IOEnclosure(Base):
    # id_field = 'id'
    _template = {'id': '',
                 'name': '',
                 'state': '',
                 }


class IOEnclosureManager(ReadOnlyManager):
    """
    Manage IO Enclosure resources.
    """
    resource_class = IOEnclosure
    resource_type = DS8K_IOENCLOSURE



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 12,
    "total_counts": 12
  },
  "data": {
    "marrays": [
      {
        "id": "MA1",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA1"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P0",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P0"
          }
        }
      },
      {
        "id": "MA2",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA2"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P1",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P1"
          }
        }
      },
      {
        "id": "MA3",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA3"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P2",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2"
          }
        }
      },
      {
        "id": "MA4",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA4"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P3",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P3"
          }
        }
      },
      {
        "id": "MA5",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA5"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P4",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P4"
          }
        }
      },
      {
        "id": "MA6",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA6"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P5",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P5"
          }
        }
      },
      {
        "id": "MA7",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA7"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P6",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P6"
          }
        }
      },
      {
        "id": "MA8",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA8"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P7",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P7"
          }
        }
      },
      {
        "id": "MA9",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA9"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P8",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P8"
          }
        }
      },
      {
        "id": "MA10",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA10"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P9",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P9"
          }
        }
      },
      {
        "id": "MA11",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA11"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P10",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P10"
          }
        }
      },
      {
        "id": "MA12",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA12"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P11",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P11"
          }
        }
      }
    ]
  }
}


ONE = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 12,
    "total_counts": 12
  },
  "data": {
    "marrays": [
      {
        "id": "MA1",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/marrays/MA1"
        },
        "disk_class": "enterprise",
        "pool": {
          "id": "P0",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P0"
          }
        }
      },
           ]
        }
       }



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
from nose.tools import nottest
from functools import cmp_to_key
import warnings
from pyds8k.exceptions import InternalServerError, FieldReadOnly
from pyds8k.messages import DEFAULT_SUCCESS_BODY_DICT
from pyds8k.resources.ds8k.v1.common.types import DS8K_HOST, \
    DS8K_VOLUME, \
    DS8K_IOPORT, \
    DS8K_HOST_PORT
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_json_by_type, \
    get_response_data_by_type
from ...data import action_response, action_response_json, \
    action_response_failed, action_response_failed_json, \
    create_host_response_json
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.ioports import IOPort
from pyds8k.resources.ds8k.v1.host_ports import HostPort
from pyds8k.resources.ds8k.v1.hosts import Host
from pyds8k.resources.ds8k.v1.volumes import Volume
from pyds8k.dataParser.ds8k import RequestParser

volume_list_response = get_response_list_data_by_type(DS8K_VOLUME)
volume_list_response_json = get_response_list_json_by_type(DS8K_VOLUME)


class TestHost(TestDS8KWithConnect):

    def test_get_volumes(self):
        self._test_sub_resource_list_by_route(DS8K_HOST, DS8K_VOLUME,
                                              self._sorted_by_volume_name
                                              )

    def test_get_ioports(self):
        self._test_sub_resource_list_by_route(
            DS8K_HOST, DS8K_IOPORT,
            self._get_sort_func_by(IOPort.id_field)
        )

    def test_get_host_ports(self):
        self._test_sub_resource_list_by_route(
            DS8K_HOST, DS8K_HOST_PORT,
            self._get_sort_func_by(HostPort.id_field)
        )

    @httpretty.activate
    def test_delete_host(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)
        httpretty.register_uri(
            httpretty.GET,
            self.domain + self.base_url + url,
            body=get_response_json_by_type(DS8K_HOST),
            content_type='application/json',
            status=200,
        )
        httpretty.register_uri(
            httpretty.DELETE,
            self.domain + self.base_url + url,
            body=action_response_json,
            content_type='application/json',
            status=204,
        )
        # Way 1
        _ = self.system.delete_host(host_name)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        # self.assertEqual(resp1, action_response['server'])

        # Way 2
        host = self.system.get_host(host_name)
        self.assertIsInstance(host, Host)
        resp2, _ = host.delete()
        self.assertEqual(resp2.status_code, 204)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        # self.assertEqual(resp2.text, action_response['server'])
        # self.assertEqual(data2, action_response['server'])
        # warnings.warn("TestHost.test_delete_host: do not know why \

    # requests can not get DELETE response's body. Maybe httpretty can \
    # not set DELETE response's body correctly")

    @httpretty.activate
    def test_delete_host_without_resp_body(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)
        httpretty.register_uri(httpretty.DELETE,
                               self.domain + self.base_url + url,
                               content_type='application/json',
                               status=204,
                               )
        resp1 = self.system.delete_host(host_name)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        self.assertEqual(resp1, DEFAULT_SUCCESS_BODY_DICT)

    @httpretty.activate
    def test_delete_host_failed(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)
        httpretty.register_uri(httpretty.DELETE,
                               self.domain + self.base_url + url,
                               body=action_response_failed_json,
                               content_type='application/json',
                               status=500,
                               )
        with self.assertRaises(InternalServerError) as cm:
            self.system.delete_host(host_name)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_update_host_rm_ioports_all(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'ioports': []})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=get_response_json_by_type(DS8K_HOST),
                               content_type='application/json',
                               status=200,
                               )
        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.update_host_rm_ioports_all(host_name)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(resp1, action_response['server'])

        host = self.system.get_host(host_name)
        # Way 2
        host.ioports = []
        resp2, data2 = host.update()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data2, action_response['server'])
        self.assertEqual(resp2.status_code, 200)

        # Way 3 in DS8K, save works the same as update
        host.ioports = []
        resp3, data3 = host.save()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data3, action_response['server'])
        self.assertEqual(resp3.status_code, 200)

        # Way 4
        host.ioports = []
        resp4, data4 = host.patch()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data4, action_response['server'])
        self.assertEqual(resp4.status_code, 200)

        # Way 5 in DS8K, put works the same as patch
        host.ioports = []
        resp5, data5 = host.put()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data5, action_response['server'])
        self.assertEqual(resp5.status_code, 200)

    @httpretty.activate
    def test_update_host_add_ioports_all(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'ioports': 'all'})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(
            httpretty.GET,
            self.domain + self.base_url + url,
            body=get_response_json_by_type(DS8K_HOST),
            content_type='application/json',
            status=200,
        )
        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.update_host_add_ioports_all(host_name)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(resp1, action_response['server'])

    @nottest
    @httpretty.activate
    def test_update_host_rm_volumes_all(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'volumes': []})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(
            httpretty.GET,
            self.domain + self.base_url + url,
            body=get_response_json_by_type(DS8K_HOST),
            content_type='application/json',
            status=200,
        )
        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.update_host_rm_volumes_all(host_name)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(resp1, action_response['server'])

        host = self.system.get_host(host_name)
        # Way 2
        host.volumes = []
        resp2, data2 = host.update()
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(data2, action_response['server'])
        self.assertEqual(resp2.status_code, 200)

    @nottest
    @httpretty.activate
    def test_update_host_add_volumes(self):
        warnings.warn('test_update_host_add_volumes: not finished yet.')

    @nottest
    @httpretty.activate
    def test_update_host_rm_volumes(self):
        warnings.warn('test_update_host_rm_volumes: not finished yet.')

    @httpretty.activate
    def test_update_host_add_ioports(self):
        response_a_json = get_response_json_by_type(DS8K_HOST)
        response_a = get_response_data_by_type(DS8K_HOST)
        host_name = self._get_resource_id_from_resopnse(DS8K_HOST, response_a,
                                                        Host.id_field
                                                        )
        res_all = get_response_list_data_by_type(DS8K_IOPORT)
        ioport_ids = self._get_resource_ids_from_resopnse(DS8K_IOPORT, res_all,
                                                          IOPort.id_field
                                                          )
        port_id = 'new_port_id'

        url = '/hosts/{}'.format(host_name)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            ioport_ids.append(port_id)
            resq = RequestParser({'ioports': ioport_ids})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        ioport_url = '{}/{}'.format(url, DS8K_IOPORT)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + ioport_url,
                               body=get_response_list_json_by_type(
                                   DS8K_IOPORT),
                               content_type='application/json',
                               status=200,
                               )
        host = self.system.get_host(host_name)
        resp = host.update_host_add_ioports(port_id)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(resp, action_response['server'])

    @httpretty.activate
    def test_update_host_rm_ioports(self):
        response_a_json = get_response_json_by_type(DS8K_HOST)
        response_a = get_response_data_by_type(DS8K_HOST)
        host_name = self._get_resource_id_from_resopnse(DS8K_HOST, response_a,
                                                        Host.id_field
                                                        )
        res_all = get_response_list_data_by_type(DS8K_IOPORT)
        ioport_ids = self._get_resource_ids_from_resopnse(DS8K_IOPORT, res_all,
                                                          IOPort.id_field
                                                          )
        port_id = ioport_ids[0]

        url = '/hosts/{}'.format(host_name)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'ioports': ioport_ids[1:]})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        ioport_url = '{}/{}'.format(url, DS8K_IOPORT)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + ioport_url,
                               body=get_response_list_json_by_type(
                                   DS8K_IOPORT),
                               content_type='application/json',
                               status=200,
                               )
        host = self.system.get_host(host_name)
        resp = host.update_host_rm_ioports(port_id)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(resp, action_response['server'])

    @httpretty.activate
    def test_update_host_failed(self):
        host_name = 'host1'
        url = '/hosts/{}'.format(host_name)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=action_response_failed_json,
                               content_type='application/json',
                               status=500
                               )
        with self.assertRaises(InternalServerError) as cm:
            self.system.update_host_rm_ioports_all(host_name)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )

    def test_set_readonly_field(self):
        host = Host(self.client)
        with self.assertRaises(FieldReadOnly):
            host.name = 'new_name'
        with self.assertRaises(FieldReadOnly):
            host.state = 'new_state'
        with self.assertRaises(FieldReadOnly):
            host.addrmode = 'new_addrmode'
        with self.assertRaises(FieldReadOnly):
            host.addrdiscovery = 'new_addrdiscovery'
        with self.assertRaises(FieldReadOnly):
            host.lbs = 'new_lbs'

    def test_set_related_resources_collection(self):
        volumes = [Volume(self.client, resource_id='volume{}'.format(i))
                   for i in range(10)
                   ]
        host_ports = \
            [HostPort(self.client, resource_id='host_port{}'.format(i))
             for i in range(10)
             ]
        ioports = [IOPort(self.client, resource_id='ioport{}'.format(i))
                   for i in range(10)
                   ]

        # init without related_resources collection
        host = Host(self.client, info={
            'volumes': 'volumes',  # string
            'ioports': '',  # empty
            'host_ports': {  # link
                'link': {
                    'rel': 'self',
                    'href': '/api/v1//host_ports'
                },
            }
        }
                    )
        for i in host.related_resources_collection:
            self.assertEqual('', host.representation.get(i))
            self.assertFalse(hasattr(host, i))

        # setting related resources collection
        for item in ((DS8K_VOLUME, volumes),
                     (DS8K_IOPORT, ioports)
                     ):
            setattr(host, item[0], item[1])
            for j, value in enumerate(host.representation[item[0]]):
                self.assertEqual(value,
                                 getattr(item[1][j], item[1][j].id_field)
                                 )
            for ind, v in enumerate(host._get_modified_info_dict()[item[0]]):
                self.assertEqual(v,
                                 getattr(item[1][ind], item[1][ind].id_field)
                                 )

        # loading related resources collection
        host.volumes = []
        host.ioports = []
        host._start_updating()
        for item in ((DS8K_VOLUME, volumes),
                     (DS8K_HOST_PORT, host_ports),
                     (DS8K_IOPORT, ioports)
                     ):
            setattr(host, item[0], item[1])
            for j, value in enumerate(host.representation[item[0]]):
                self.assertEqual(value,
                                 getattr(item[1][j], item[1][j].id_field)
                                 )
        host._stop_updating()

    @httpretty.activate
    def test_lazy_loading_related_resources_collection(self):
        response_a_json = get_response_json_by_type(DS8K_HOST)
        response_a = get_response_data_by_type(DS8K_HOST)
        host_name = self._get_resource_id_from_resopnse(DS8K_HOST, response_a,
                                                        Host.id_field
                                                        )
        url = '/hosts/{}'.format(host_name)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )
        for item in Host.related_resources_collection:
            sub_route_url = '{}/{}'.format(url, item)
            httpretty.register_uri(httpretty.GET,
                                   self.domain + self.base_url + sub_route_url,
                                   body=get_response_list_json_by_type(item),
                                   content_type='application/json',
                                   status=200,
                                   )
        host = self.system.get_host(host_name)

        for item in Host.related_resources_collection:
            res_collection = getattr(host, item)
            self.assertNotEqual(0, len(res_collection))
            res_collection.sort(
                key=cmp_to_key(
                    self._get_sort_func_by(res_collection[0].id_field))
            )
            res_collection_data = list(
                get_response_list_data_by_type(item)['data'][item]
            )
            res_collection_data.sort(
                key=cmp_to_key(
                    self._get_sort_func_by(res_collection[0].id_field))
            )

            self.assertEqual(
                len(res_collection_data),
                len(res_collection))
            self._assert_equal_between_sorted_dict_and_resource_list(
                res_collection_data,
                res_collection
            )

    def test_set_related_resources_collection_during_loading(self):
        host = Host(self.client, info={
            'volumes': [{
                'id': '0000',
                'link': {
                    'rel': 'self',
                    'href': '/api/v1/volumes/0000'
                },
            },
            ],
            'ioports': [{
                'id': '0030',
                'link': {
                    'rel': 'self',
                    'href': '/api/v1/ioports/0030'
                },
            },
            ],  # missing
            'host_ports': [{
                'wwpn': '50050763030313A2',
                'link': {
                    'rel': 'self',
                    'href': '/api/v1/host_ports/50050763030313A2'
                },
            },
            ],
        }
                    )

        self.assertEqual('0000', host.representation.get('volumes')[0])
        self.assertEqual('0030', host.representation.get('ioports')[0])
        self.assertEqual('50050763030313A2',
                         host.representation.get('host_ports')[0]
                         )
        self.assertEqual('0000', host.volumes[0].id)
        self.assertEqual('0030', host.ioports[0].id)
        self.assertEqual('50050763030313A2', host.host_ports[0].id)

    @httpretty.activate
    def test_create_host(self):
        host_type = 'VMware'
        url = '/hosts'
        host_name = 'host1'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            req = RequestParser({'hosttype': host_type, 'name': host_name})
            self.assertDictContainsSubset(
                req.get_request_data().get('request').get('params'),
                json.loads(request.body).get('request').get('params'),
            )
            return (200, headers, create_host_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.system.create_host(host_name, host_type)
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp1[0], Host)

        # Way 2
        host = self.system.all(DS8K_HOST, rebuild_url=True)
        new_host2 = host.create(hosttype=host_type, name=host_name)
        resp2, data2 = new_host2.posta()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data2[0], Host)
        self.assertEqual(resp2.status_code, 200)

        # Way 3
        host = self.system.all(DS8K_HOST, rebuild_url=True)
        new_host3 = host.create(hosttype=host_type, name=host_name)
        resp3, data3 = new_host3.save()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data3[0], Host)
        self.assertEqual(resp3.status_code, 200)

        # Way 4
        # Don't init a resource instance by yourself when create new.
        # use .create() instead.

    @httpretty.activate
    def test_create_host_failed(self):
        host_type = 'VMware'
        url = '/hosts'
        host_name = 'host1'

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=action_response_failed_json,
                               content_type='application/json',
                               status=500
                               )
        with self.assertRaises(InternalServerError) as cm:
            self.system.create_host(host_name, host_type)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.exceptions import URLParseError
from . import base
import httpretty
import json
import time
from nose.tools import nottest
from pyds8k.httpclient import HTTPClient
from pyds8k.base import Resource, DefaultManager
from .data import get_response_list_json_by_type, \
                      get_response_list_data_by_type, \
                      get_response_data_by_type, \
                      get_response_json_by_type
info = {'id': 'v1', 'name': 'vol1'}

custom_method_get = {'data': 'custom_method_get'}
custom_method_get_json = json.dumps(custom_method_get)

DEFAULT = 'default'
default_a_response = get_response_data_by_type(DEFAULT)
default_a_response_json = get_response_json_by_type(DEFAULT)
default_list_response = get_response_list_data_by_type(DEFAULT)
default_list_response_json = get_response_list_json_by_type(DEFAULT)


class TestHTTPClient(base.TestCaseWithConnect):

    def setUp(self):
        super(TestHTTPClient, self).setUp()
        self.base_url = ''

    def test_parse_url(self):
        url1 = self.domain + '/new'
        url2 = '/new'
        _, url3 = url1.split('//')
        url4 = 'http://new_domain' + '/new'
        self.assertEqual('/new', self.client._parse_url(url1))
        self.assertEqual('/new', self.client._parse_url(url2))
        self.assertEqual('/new', self.client._parse_url(url3))
        with self.assertRaises(URLParseError):
            self.client._parse_url(url4)
        new_client = HTTPClient('9.115.247.115', 'admin', 'admin',
                                service_type='ds8k',
                                secure=True)
        with self.assertRaises(URLParseError):
            new_client._parse_url(url3)

    @httpretty.activate
    def test_redirect(self):
        url = '/default/old'
        new_url = '/default/a'
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               content_type='application/json',
                               adding_headers={'Location': new_url},
                               status=301)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + new_url,
                               body=default_a_response_json,
                               content_type='application/json',
                               status=200)
        de = self.resource.one(DEFAULT, 'old').get(allow_redirects=False)
        self.assertEqual(new_url, de.url)

    # Not work in this way.
    @nottest
    @httpretty.activate
    def test_timeout(self):
        url = '/default/a'
        new_client = HTTPClient('localhost', 'admin', 'admin',
                                service_type='ds8k',
                                timeout=0.01)

        def _verify_request(request, uri, headers):
            time.sleep(10)
            return (200, headers, default_a_response_json)

        httpretty.register_uri(httpretty.GET,
                               new_client.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )

        resource = Resource(new_client, DefaultManager(new_client))
        resource.one(DEFAULT, 'a').get()



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
   "server":
   {
       "status": "ok",
       "code": "",
       "message": "Operation done successfully."
   },
   "counts":
   {
       "data_counts": 19231,
       "total_counts": 19231
   },
   "data":
   {
       "events":
       [
           {
               "id": "SE1",
               "type": "UserLoginFailed",
               "severity": "info",
               "time": "2015-03-31T08:00:48-0700",
               "resource_id": "EHqqq92V9",
               "formatted_parameter":
               [
                   "EHqqq92V9",
                   "(2) the use of an account that does not exist",
                   "lockedfalse",
                   "1",
                   "initialPolicy"
               ],
               "description": "event details",
           },
           {
               "id": "SE2",
               "type": "UserLoginFailed",
               "severity": "info",
               "time": "2015-03-31T08:00:48-0700",
               "resource_id": "EHqqq92V9",
               "formatted_parameter":
               [
                   "EHqqq92V9",
                   "(2) the use of an account that does not exist",
                   "lockedfalse",
                   "1",
                   "initialPolicy"
               ],
               "description": "event details",
           },
           {
               "id": "SE3",
               "type": "UserLoggedOn",
               "severity": "info",
               "time": "2015-03-31T08:00:48-0700",
               "resource_id": "admin",
               "formatted_parameter":
               [
                   "admin",
                   ",Administrator",
                   "*",
                   "9.11.217.179",
                   "DSGUI",
                   "5.7.40.1303",
                   "initialPolicy",
                   "",
                   "HMCID: 1",
                   "fbb1149"
               ],
               "description": "event details",
           },
           {
               "id": "SE4",
               "type": "UserLoggedOn",
               "severity": "info",
               "time": "2015-03-31T08:00:48-0700",
               "resource_id": "admin",
               "formatted_parameter":
               [
                   "admin",
                   ",Administrator",
                   "*",
                   "9.11.217.179",
                   "DSGUI",
                   "87.40.141.0",
                   "initialPolicy",
                   "9.123.236.47",
                   "HMCID: 1",
                   "59d2965a"
               ],
               "description": "event details",
           },
        ]
    }
}


ONE = {
   "server":
   {
       "status": "ok",
       "code": "",
       "message": "Operation done successfully."
   },
   "counts":
   {
       "data_counts": 1,
       "total_counts": 1
   },
   "data":
   {
       "events":
       [
           {
               "id": "SE1",
               "type": "UserLoginFailed",
               "severity": "info",
               "time": "2015-03-31T08:00:48-0700",
               "resource_id": "EHqqq92V9",
               "formatted_parameter":
               [
                   "EHqqq92V9",
                   "(2) the use of an account that does not exist",
                   "lockedfalse",
                   "1",
                   "initialPolicy"
               ],
               "description": "event details",
           },
        ]
    }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from functools import wraps, partial
from logging import getLogger
from contextlib import contextmanager
import unittest
from nose.tools import nottest
from datetime import datetime
from pyds8k import PYDS8K_DEFAULT_LOGGER
from pyds8k.utils import res_timer_recorder
from pyds8k.client.ds8k.v1.client import Client
from pyds8k.client.ds8k.v1.sc_client import SCClient
from pyds8k.resources.ds8k.v1.common import types
from pyds8k.size_converter import convert_size_gib_to_bytes

from .device import ds44 as ds8k_device

logger = getLogger(PYDS8K_DEFAULT_LOGGER)


def add_logger(route):
    def add_logger_deco(func):
        @wraps(func)
        def inner(self, route_id=None):
            if route_id:
                logger.info(
                    'Starting GET /{}/{} request'.format(route,
                                                         route_id
                                                         )
                    )
                res = func(self, route_id)
                logger.info(
                    'Successfully got {}: {}, detail is: {}'.format(
                        route,
                        res,
                        res.representation
                        )
                    )
                logger.info('Finish GET /{}/{} request'.format(route, res.id))
            else:
                logger.info('Starting GET /{} request'.format(route))
                res = func(self)
                logger.info(
                    'Successfully got {} {}: {}'.format(len(res),
                                                        route,
                                                        res,
                                                        )
                    )
                logger.info('Finish GET /{} request'.format(route))
            return res
        return inner
    return add_logger_deco


class TestIntegration(unittest.TestCase):

    @classmethod
    def setup_class(cls):

        cls.client = Client(ds8k_device.ipaddr,
                            ds8k_device.username, ds8k_device.password,
                            # hostname='mtc032h.tuc.stglabs.ibm.com',
                            port=ds8k_device.port,
                            )

    def __getattr__(self, k):
        if k.startswith('get_'):
            _, route = str(k).split('_', 1)

            @add_logger(route=route)
            def route_getter(self, route_id=None):
                return getattr(self.client, k)(route_id)
            return partial(route_getter, self)

        return super(TestIntegration, self).__getattr__(k)

    def test_system(self):
        logger.info('Starting GET /systems request')
        sys = self.client.get_system()
        logger.info(
            'Successfully got system: {}, detail is: {}'.format(
                sys,
                sys.representation
                )
            )
        logger.info('Finish GET /systems request')

    def test_nodes(self):
        nodes = self.get_nodes()
        self.get_nodes(nodes[0].id)

    def test_marrays(self):
        marrays = self.get_marrays()
        self.get_marrays(marrays[0].id)

    def test_lss(self):
        lss_list = self.get_lss()
        for l in lss_list:
            lss = self.get_lss(l.id)
            volumes = self._get_volumes_by(types.DS8K_LSS, lss)
            if volumes:
                self.get_volumes(volumes[0].id)
                return

    def test_ioports(self):
        ios = self.get_ioports()
        self.get_ioports(ios[0].id)

    def test_flashcopy(self):
        self.get_flashcopy()

        volume = self.get_volumes('0000')
        self._get_sub_resource_by(types.DS8K_VOLUME, volume,
                                  types.DS8K_FLASHCOPY
                                  )

    def test_pprc(self):
        self.get_pprc()

        volume = self.get_volumes('0000')
        self._get_sub_resource_by(types.DS8K_VOLUME, volume,
                                  types.DS8K_PPRC
                                  )

    def test_events(self):
        sys = self.client.get_system()
        before = datetime.now()
        after = datetime(year=before.year,
                         month=before.month,
                         day=before.day)
        logger.info('Starting GET /events request')
        events = sys.get_events_by_filter(warning=True,
                                          error=True,
                                          before=before,
                                          after=after)
        logger.info(
            'Successfully got {} events'.format(len(events))
            )
        logger.info('Finish GET /events request')

    def test_pools(self):
        pools = self.get_pools()
        for p in pools:
            pool = self.get_pools(p.id)
            vols = self._get_volumes_by(types.DS8K_POOL, pool)
            if vols:
                return

    # Skipped, because get all volumes is not allowed.
    @nottest
    def test_volumes(self):
        volumes = self.get_volumes()
        self.get_volumes(volumes[0].id)

    def test_volume(self):
        self.get_volumes('0000')

    def test_hosts(self):
        hosts = self.get_hosts()
        for h in hosts:
            host = self.get_hosts(h.id)
            volumes = self._get_sub_resource_by(types.DS8K_HOST, host,
                                                types.DS8K_VOLUME
                                                )
            mappings = self._get_sub_resource_by(types.DS8K_HOST, host,
                                                 types.DS8K_VOLMAP
                                                 )
            ioports = self._get_sub_resource_by(types.DS8K_HOST, host,
                                                types.DS8K_IOPORT
                                                )
            host_ports = self._get_sub_resource_by(types.DS8K_HOST, host,
                                                   types.DS8K_HOST_PORT
                                                   )
            if volumes and mappings and ioports and host_ports:
                return

    def test_host_ports(self):
        ports = self.get_host_ports()
        if ports:
            self.get_host_ports(ports[0].id)

    def _get_volumes_by(self, route, parent_res):
        return self._get_sub_resource_by(route, parent_res, types.DS8K_VOLUME)

    @res_timer_recorder
    def _get_sub_resource_by(self, route, parent_res, sub_route):
        logger.info('Starting GET /{}/{}/{} request'.format(route,
                                                            parent_res.id,
                                                            sub_route
                                                            )
                    )
        # Lazy-loading
        sub_res = getattr(parent_res, sub_route)
        logger.info('Successfully got {} {}'.format(len(sub_res), sub_route))
        logger.info('Finish GET /{}/{}/{} request'.format(route,
                                                          parent_res.id,
                                                          sub_route
                                                          )
                    )
        return sub_res


class TestSCClient(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        cls.client = SCClient(ds8k_device.ipaddr,
                              ds8k_device.username, ds8k_device.password,
                              # hostname='mtc032h.tuc.stglabs.ibm.com',
                              port=ds8k_device.port,
                              )

    def test_volume_create_and_delete(self):
        volume = self._prepare_volume()
        self._destroy_volume(volume.id)

    def test_volume_rename(self):
        new_name = 'test_rename'
        with self.get_test_volume() as volume:
            res = self.client.rename_volume(volume.id, new_name)
            new_volume = self.client.get_volume(volume.id)[0]
            self.assertEqual(new_volume.get('name'), new_name)
            logger.info(
                'Successfully renamed the volume {}, response is {}'.format(
                    volume.id, res
                    )
                )

    def test_volume_extend(self):
        new_size = '7'
        with self.get_test_volume() as volume:
            res = self.client.extend_volume(volume.id, new_size)
            new_volume = self.client.get_volume(volume.id)[0]
            self.assertEqual(new_volume.get('cap'),
                             str(convert_size_gib_to_bytes(int(new_size)))
                             )
            logger.info(
                'Successfully extended the volume {}, response is {}'.format(
                    volume.id, res
                    )
            )

    def test_volume_move(self):
        with self.get_test_volume() as volume:
            pools = self.client.list_extentpools()
            old_pool_id = volume.pool
            for pool in pools:
                if pool.get('id') != old_pool_id:
                    new_pool_id = pool.get('id')
                    res = self.client.relocate_volume(volume.id, new_pool_id)
                    new_volume = self.client.get_volume(volume.id)[0]
                    self.assertEqual(new_volume.get('pool'),
                                     new_pool_id
                                     )
                    logger.info(
                        'Successfully move volume {} from pool {} to pool {}, response is {}'.format(  # noqa
                            volume.id, old_pool_id, new_pool_id, res
                            )
                        )
                    break

    def test_host_create_and_delete(self):
        host_name = self._prepare_host()
        self._destroy_host(host_name)

    def test_volume_map_and_unmap(self):
        with self.get_test_host() as host_name:
            with self.get_test_volume() as volume:
                used_lunids = self.client.get_used_lun_numbers_by_host(
                    host_name
                )
                unused_lunids = \
                    ['{0:0{1}x}'.format(i, 2) for i in range(256)
                     if '{0:0{1}x}'.format(i, 2) not in used_lunids
                     ]
                lunid = unused_lunids[0]
                logger.info(
                    'Trying to map volume {} to host {} with lunid {}'.format(
                        volume.id, host_name, lunid
                    )
                )
                res = self.client.map_volume_to_host(host_name=host_name,
                                                     volume_id=volume.id,
                                                     lunid=lunid
                                                     )
                logger.info(
                    'Successfully map volume {} to host {}. res is {}'.format(
                        volume.id, host_name, res
                    )
                )
                logger.info(
                    'Trying to unmap volume {} from host {}.'.format(
                        volume.id, host_name
                    )
                )
                res = self.client.unmap_volume_from_host(host_name, lunid)
                logger.info(
                    'Successfully unmap volume {} from host {}. res is {}'.format(  # noqa
                        volume.id, host_name, res
                    )
                )

    def test_volume_map_and_unmap_to_zlinux_type_host(self):
        with self.get_test_zlinux_type_host() as host_name:
            with self.get_test_volume() as volume:
                logger.info(
                    'Trying to map volume {} to host {}'.format(
                        volume.id, host_name
                    )
                )
                res = self.client.map_volume_to_host(host_name=host_name,
                                                     volume_id=volume.id,
                                                     lunid=''
                                                     )
                logger.info(
                    'Successfully map volume {} to host {}. res is {}'.format(
                        volume.id, host_name, res
                    )
                )
                logger.info(
                    'Trying to unmap volume {} from host {}.'.format(
                        volume.id, host_name
                    )
                )
                lunid = int('40' + volume.id[:2] + '40' + volume.id[2:], 16)
                res = self.client.unmap_volume_from_host(host_name, lunid)
                logger.info(
                    'Successfully unmap volume {} from host {}. res is {}'.format(  # noqa
                        volume.id, host_name, res
                    )
                )

    def _prepare_volume(self):
        logger.info('Preparing a new volume for test purpose.')
        pools = self.client.list_extentpools()
        res = self.client.create_volumes(pool_id=pools[0].get('id'),
                                         capacity_in_GiB=2,
                                         sam='ese',
                                         volume_names_list=['loutest_volume1']
                                         )
        logger.info('Task done, the volume {} is created.'.format(res))
        return res[0]

    def _destroy_volume(self, volume_id):
        logger.info('Destroying the created volume {}'.format(volume_id))
        # delete may fail if it is mapped to hosts.
        self.client.delete_volume(volume_id)
        logger.info('Task done, the volume is deleted successfully.')

    def _prepare_host(self):
        logger.info('Preparing a new host for test purpose.')
        res = self.client.crate_host(host_name='loutest_host1',
                                     wwpn='1'
                                     )
        logger.info('Task done, the host {} is created.'.format(res))
        return res

    def _prepare_host_of_zlinux(self):
        logger.info('Preparing a new zLinux type host for test purpose.')
        res = self.client.crate_host(host_name='zlinuxtest_host1',
                                     wwpn='1',
                                     host_type='zLinux'
                                     )
        logger.info('Task done, the zLinux type host {} is '
                    'created.'.format(res))
        return res

    def _destroy_host(self, host_name):
        logger.info('Destroying the created host {}'.format(host_name))
        self.client.delete_host(host_name)
        logger.info('Task done, the host is deleted successfully.')

    @contextmanager
    def get_test_volume(self):
        volume = self._prepare_volume()
        try:
            yield volume
        except Exception:
            raise
        finally:
            self._destroy_volume(volume.id)

    @contextmanager
    def get_test_host(self):
        host_name = self._prepare_host()
        try:
            yield host_name
        except Exception:
            raise
        finally:
            self._destroy_host(host_name)

    @contextmanager
    def get_test_zlinux_type_host(self):
        host_name = self._prepare_host_of_zlinux()
        try:
            yield host_name
        except Exception:
            raise
        finally:
            self._destroy_host(host_name)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.dataParser.ds8k import RequestParser, ResponseParser

AUTH_URL = '/tokens'
DEFAULT_BASE_URL = '/api/v1'


class Auth(object):

    base_url = DEFAULT_BASE_URL
    auth_url = AUTH_URL

    def __init__(self):
        pass

    @classmethod
    def authenticate(self, client):
        """
        The main authenticate method. Mandatory
        """

        params = {}
        params['username'] = client.user
        params['password'] = client.password
        if client.hostname:
            params['hmc1'] = client.hostname
        req_p = RequestParser(params)
        _, body = client.post(self.get_auth_url(),
                              body=req_p.get_request_data()
                              )
        token = _get_data(body).get('token', '')
        if token:
            client.set_defaultHeaders('X-Auth-Token', token)
            # client.set_defaultQuerystrings('token', token)

    @classmethod
    def get_auth_url(self):
        """
        Return the auth url. Mandatory
        """

        return self.base_url + self.auth_url


def _get_data(response_body):
    res_p = ResponseParser(response_body, 'token')
    return res_p.get_representations()[0]



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from io import StringIO
import sys
from contextlib import contextmanager


@contextmanager
def capture_sys_stderr_and_return(command, *args, **kwargs):
    err, sys.stderr = sys.stderr, StringIO()
    command(*args, **kwargs)
    sys.stderr.seek(0)
    yield sys.stderr.read()
    sys.stderr = err



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from abc import ABCMeta, abstractmethod


class BaseRequestParser(object, metaclass=ABCMeta):
    """
    Parse the data user wants to send to server,
    in the right format that server defined.
    """

    @abstractmethod
    def get_raw_data(self):
        pass

    @abstractmethod
    def get_request_data(self):
        pass


class BaseResponseParser(object, metaclass=ABCMeta):
    """
    Parser response data, to get resource link, representation, etc.
    """

    response_key = ''
    success_status = ('ok', 'successful', )
    resource_data_key = 'data'
    error_status_key = 'status'

    @abstractmethod
    def get_raw_data(self):
        pass

    @abstractmethod
    def get_link(self):
        pass

    @abstractmethod
    def get_representations(self):
        pass

    @abstractmethod
    def get_posta_response_data(self):
        """
        Return a list of ({resource_data_key:resource_data}, resource_uri)
        or ({error_status_key:status_data}, None) if the rest server returns
        a partial failure response.

        For example:
        If you create two resources(res1, res2) at a time, you should return
        [({resource_data_key:res1_data}, res1_uri),
         ({error_status_key:res2_status_data}, None)
         ]
        or
        [({resource_data_key:res1_data}, res1_uri), ] if you only care about
        success ones.
        """
        pass

    @abstractmethod
    def get_status_body(self):
        pass

    @abstractmethod
    def get_error_code(self):
        pass

    @abstractmethod
    def get_error_msg(self):
        pass

    @classmethod
    @abstractmethod
    def get_link_from_representation(cls, representation):
        pass



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 32,
    "total_counts": 32
  },
  "data": {
    "ioports": [
      {
        "id": "I0200",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/ioports/I0200"
        },
        "state": "Online",
        "protocol": "SCSI-FCP",
        "wwpn": "500507630310003D",
        "type": "Fibre Channel-SW",
        "speed": "8 Gb/s",
        "loc": "U1400.1B3.44001B3-P1-C1-T0",
        "io_enclosure": {
            "id": "2",
            "link": {
                "rel": "self",
                "href": "https://localhost:8088/api/v1/io_enclosures/2"
                },
        }
      },
      {
        "id": "I0201",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/ioports/I0201"
        },
        "state": "Online",
        "protocol": "SCSI-FCP",
        "wwpn": "500507630310403D",
        "type": "Fibre Channel-SW",
        "speed": "8 Gb/s",
        "loc": "U1400.1B3.44001B3-P1-C1-T1",
        "io_enclosure": {
            "id": "2",
            "link": {
                "rel": "self",
                "href": "https://localhost:8088/api/v1/io_enclosures/2"
                },
        }
      },
      {
        "id": "I0202",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/ioports/I0202"
        },
        "state": "Online",
        "protocol": "SCSI-FCP",
        "wwpn": "500507630310803D",
        "type": "Fibre Channel-SW",
        "speed": "8 Gb/s",
        "loc": "U1400.1B3.44001B3-P1-C1-T2",
        "io_enclosure": {
            "id": "2",
            "link": {
                "rel": "self",
                "href": "https://localhost:8088/api/v1/io_enclosures/2"
                },
        }
      },
      {
        "id": "I0336",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/ioports/I0336"
        },
        "state": "Offline",
        "protocol": "SCSI-FCP",
        "wwpn": "50050763035B803D",
        "type": "Fibre Channel-LW",
        "speed": "8 Gb/s",
        "loc": "U1400.1B4.44001B4-P1-C4-T6",
        "io_enclosure": {
            "id": "2",
            "link": {
                "rel": "self",
                "href": "https://localhost:8088/api/v1/io_enclosures/2"
                },
        }
      },
      {
        "id": "I0337",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/ioports/I0337"
        },
        "state": "Offline",
        "protocol": "SCSI-FCP",
        "wwpn": "50050763035BC03D",
        "type": "Fibre Channel-LW",
        "speed": "8 Gb/s",
        "loc": "U1400.1B4.44001B4-P1-C4-T7",
        "io_enclosure": {
            "id": "2",
            "link": {
                "rel": "self",
                "href": "https://localhost:8088/api/v1/io_enclosures/2"
                },
        }
      }
    ]
  }
}
ONE = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 1,
    "total_counts": 1
  },
  "data": {
    "ioports": [
      {
        "id": "I0200",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/ioports/I0200"
        },
        "state": "Online",
        "protocol": "SCSI-FCP",
        "wwpn": "500507630310003D",
        "type": "Fibre Channel-SW",
        "speed": "8 Gb/s",
        "loc": "U1400.1B3.44001B3-P1-C1-T0",
        "io_enclosure": {
            "id": "2",
            "link": {
                "rel": "self",
                "href": "https://localhost:8088/api/v1/io_enclosures/2"
                },
        }
      }
    ]
  }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.resources.ds8k.v1.common.types import DS8K_SYSTEM, \
    DS8K_VOLUME
from ...base import TestCaseWithConnect
from pyds8k.resources.ds8k.v1.common.base import Base
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type
from pyds8k.client.ds8k.v1.client import Client
from pyds8k.resources.ds8k.v1.volumes import Volume
import httpretty

system_list_response_json = get_response_list_json_by_type(DS8K_SYSTEM)
volume_list_response_json = get_response_list_json_by_type(DS8K_VOLUME)
volume_list_response = get_response_list_data_by_type(DS8K_VOLUME)


class TestClient(TestCaseWithConnect):

    def setUp(self):
        super(TestClient, self).setUp()
        self.base_url = Base.base_url
        self.rest_client = Client('localhost', 'admin', 'admin', '8088')

    @httpretty.activate
    def test_get_array_method(self):
        domain = self.client.domain
        vol_url = '/volumes'
        sys_url = '/systems'
        httpretty.register_uri(httpretty.GET,
                               domain + self.base_url + vol_url,
                               body=volume_list_response_json,
                               content_type='application/json')

        httpretty.register_uri(httpretty.GET,
                               domain + self.base_url + sys_url,
                               body=system_list_response_json,
                               content_type='application/json')

        vol_list = self.rest_client.get_volumes()
        self.assertIsInstance(vol_list, list)
        self.assertIsInstance(vol_list[0], Volume)
        self.assertEqual(
                         len(vol_list),
                         len(volume_list_response['data']['volumes'])
                         )
        with self.assertRaises(AttributeError):
            # 'base_url' is an attr from System, not a method
            self.rest_client.base_url



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
    'data': {
        'default': [
            {
                'id': 'a',
                'link': {
                    'rel': 'self',
                    'href': '/default/a'
                },
            },
            {
                'id': 'b',
                'link': {
                    'rel': 'self',
                    'href': '/default/b'
                },
            }
        ]
    }
}

ONE = {
    'data': {
        'default': [
            {'name': 'name_a',
             'id': 'a',
             'link': {
                 'rel': 'self',
                 'href': '/default/a'
             },
             }
        ]
    }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 1,
    "total_counts": 1
  },
  "data": {
    "systems": [
      {
        "id": "2107-75DHZ81",
        "name": "mtc032h",
        "state": "online",
        "release": "7.4",
        "bundle": "87.40.47.0",
        "MTM": "2421-961",
        "sn": "75DHZ81",
        "wwnn": "5005076306FFD2F0",
        "cap": "440659",
        "capalloc": "304361",
        "capavail": "136810",
        "capraw": "73282879488"
      }
    ]
  }
}

ONE = ALL



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 32,
    "total_counts": 32
  },
  'data': {
    'pprc': [{
        'sourcevolume': {
            'id': '0000',
            'link': {
                'rel': 'self',
                'href': 'https://localhost:8088/api/v1/volumes/0000',
            },
        },
        'targetvolume': {
            'id': '0001',
            'link': {
                'rel': 'self',
                'href': 'https://localhost:8088/api/v1/volumes/0001',
            },
        },
        'targetsystem': {
            'id': 'remote_ds8k',
            'link': {},
        },
        'type': 'globalcopy',
        'state': 'copy_pending',
    }, {
        'sourcevolume': {
            'id': '1000',
            'link': {
                'rel': 'self',
                'href': 'https://localhost:8088/api/v1/volumes/1000',
            },
        },
        'targetvolume': {
            'id': '1001',
            'link': {
                'rel': 'self',
                'href': 'https://localhost:8088/api/v1/volumes/1001',
            },
        },
        'targetsystem': {
            'id': 'remote_ds8k',
            'link': {},
        },
        'type': 'globalcopy',
        'state': 'copy_pending',
    },
    ]
  }
}

ONE = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 1,
    "total_counts": 1
  },
  'data': {
      'pprc': [
          {
              'sourcevolume': {
                  'id': '0000',
                  'link': {
                      'rel': 'self',
                      'href': 'https://localhost:8088/api/v1/volumes/0000',
                  },
              },
              'targetvolume': {
                  'id': '0001',
                  'link': {
                      'rel': 'self',
                      'href': 'https://localhost:8088/api/v1/volumes/0001',
                  },
              },
              'targetsystem': {
                  'id': 'remote_ds8k',
                  'link': {},
              },
              'type': 'globalcopy',
              'state': 'copy_pending',
          },
      ]
  }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
from nose.tools import nottest
from pyds8k.resources.ds8k.v1.common import types
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_data_by_type, \
    get_response_json_by_type
from ...data import action_response_json
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.systems import System
from pyds8k.resources.ds8k.v1.lss import LSS
# from pyds8k.resources.ds8k.v1.ioports import IOPort
from pyds8k.resources.ds8k.v1.tserep import TSERep
from pyds8k.resources.ds8k.v1.eserep import ESERep

system_list_response = get_response_list_data_by_type(types.DS8K_SYSTEM)
system_list_response_json = get_response_list_json_by_type(types.DS8K_SYSTEM)
lss_list_response = get_response_list_data_by_type(types.DS8K_LSS)
lss_list_response_json = get_response_list_json_by_type(types.DS8K_LSS)
lss_a_response = get_response_data_by_type(types.DS8K_LSS)
lss_a_response_json = get_response_json_by_type(types.DS8K_LSS)
ioport_list_response = get_response_list_data_by_type(types.DS8K_IOPORT)
ioport_list_response_json = get_response_list_json_by_type(types.DS8K_IOPORT)
ioport_a_response = get_response_data_by_type(types.DS8K_IOPORT)
ioport_a_response_json = get_response_json_by_type(types.DS8K_IOPORT)
tserep_list_response_json = get_response_list_json_by_type(types.DS8K_TSEREP)
eserep_list_response_json = get_response_list_json_by_type(types.DS8K_ESEREP)


class TestRootResourceMixin(TestDS8KWithConnect):

    @httpretty.activate
    def test_get_system(self):
        url = '/systems'
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=system_list_response_json,
                               content_type='application/json',
                               status=200,
                               )
        sys = self.system.get_system()
        self.assertIsInstance(sys, System)
        sys_data = system_list_response['data']['systems'][0]
        self._assert_equal_between_dict_and_resource(sys_data, sys)

    def test_get_lss(self):
        self._test_resource_list_by_route(types.DS8K_LSS)

    def test_get_fb_lss(self):
        self._test_get_lss_by_type(types.DS8K_VOLUME_TYPE_FB)

    def test_get_ckd_lss(self):
        self._test_get_lss_by_type(types.DS8K_VOLUME_TYPE_CKD)

    @httpretty.activate
    def _test_get_lss_by_type(self, lss_type='fb'):
        url = '/lss'
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=lss_list_response_json,
                               content_type='application/json',
                               status=200,
                               )
        self.system.get_lss(lss_type=lss_type)
        self.assertEqual([lss_type, ],
                         httpretty.last_request().querystring.get('type')
                         )

    @httpretty.activate
    def test_get_lss_by_id(self):
        lss_id = '00'
        url = '/lss/{}'.format(lss_id)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=lss_a_response_json,
                               content_type='application/json',
                               status=200,
                               )
        lss = self.system.get_lss_by_id(lss_id)
        self.assertIsInstance(lss, LSS)
        lss_data = lss_a_response['data']['lss'][0]
        self._assert_equal_between_dict_and_resource(lss_data, lss)

    def test_get_ioports(self):
        self._test_resource_list_by_route(types.DS8K_IOPORT)

    def test_get_ioport(self):
        self._test_resource_by_route(types.DS8K_IOPORT)

    def test_get_host_ports(self):
        self._test_resource_list_by_route(types.DS8K_HOST_PORT)

    def test_get_host_port(self):
        self._test_resource_by_route(types.DS8K_HOST_PORT)

    def test_get_hosts(self):
        self._test_resource_list_by_route(types.DS8K_HOST)

    def test_get_host(self):
        self._test_resource_by_route(types.DS8K_HOST)

    def test_get_pools(self):
        self._test_resource_list_by_route(types.DS8K_POOL)

    def test_get_pool(self):
        self._test_resource_by_route(types.DS8K_POOL)

    def test_get_nodes(self):
        self._test_resource_list_by_route(types.DS8K_NODE)

    def test_get_node(self):
        self._test_resource_by_route(types.DS8K_NODE)

    def test_get_marrays(self):
        self._test_resource_list_by_route(types.DS8K_MARRAY)

    def test_get_marray(self):
        self._test_resource_by_route(types.DS8K_MARRAY)

    def test_get_users(self):
        self._test_resource_list_by_route(types.DS8K_USER)

    def test_get_user(self):
        self._test_resource_by_route(types.DS8K_USER)

    def test_get_io_enclosures(self):
        self._test_resource_list_by_route(types.DS8K_IOENCLOSURE)

    def test_get_io_enclosure(self):
        self._test_resource_by_route(types.DS8K_IOENCLOSURE)

    def test_get_encryption_groups(self):
        self._test_resource_list_by_route(types.DS8K_ENCRYPTION_GROUP)

    def test_get_encryption_group(self):
        self._test_resource_by_route(types.DS8K_ENCRYPTION_GROUP)

    def test_get_flashcopies(self):
        self._test_resource_list_by_route(types.DS8K_FLASHCOPY)

    @nottest
    def test_get_flashcopy(self):
        self._test_resource_by_route(types.DS8K_FLASHCOPY)

    def test_get_pprc(self):
        self._test_resource_list_by_route(types.DS8K_PPRC)

    @nottest
    def test_get_pprc_by_id(self):
        self._test_resource_by_route(types.DS8K_PPRC)

    def test_get_events(self):
        self._test_resource_list_by_route(types.DS8K_EVENT)

    def test_get_event(self):
        self._test_resource_by_route(types.DS8K_EVENT)

    @httpretty.activate
    def test_delete_tserep_by_pool(self):
        pool_name = 'testpool_0'
        url = '/pools/{}/tserep'.format(pool_name)
        httpretty.register_uri(httpretty.DELETE,
                               self.domain + self.base_url + url,
                               body=action_response_json,
                               content_type='application/json',
                               status=204,
                               )
        self.system.delete_tserep_by_pool(pool_name)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_delete_eserep_by_pool(self):
        pool_name = 'testpool_0'
        url = '/pools/{}/eserep'.format(pool_name)
        httpretty.register_uri(
                               httpretty.DELETE,
                               self.domain + self.base_url + url,
                               body=action_response_json,
                               content_type='application/json',
                               status=204,
                               )
        self.system.delete_eserep_by_pool(pool_name)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_get_tserep_by_pool(self):
        pool_name = 'testpool_0'
        url = '/pools/{}/tserep'.format(pool_name)
        httpretty.register_uri(
                               httpretty.GET,
                               self.domain + self.base_url + url,
                               body=tserep_list_response_json,
                               content_type='application/json',
                               status=200,
                               )
        tserep = self.system.get_tserep_by_pool(pool_name)
        self.assertIsInstance(tserep, TSERep)

    @httpretty.activate
    def test_get_eserep_by_pool(self):
        pool_name = 'testpool_0'
        url = '/pools/{}/eserep'.format(pool_name)
        httpretty.register_uri(
                               httpretty.GET,
                               self.domain + self.base_url + url,
                               body=eserep_list_response_json,
                               content_type='application/json',
                               status=200,
                               )
        eserep = self.system.get_eserep_by_pool(pool_name)
        self.assertIsInstance(eserep, ESERep)

    def test_get_volumes(self):
        self._test_resource_list_by_route(types.DS8K_VOLUME)

    def test_get_volume(self):
        self._test_resource_by_route(types.DS8K_VOLUME)

    def test_get_volumes_by_lss(self):
        pass

    def test_get_volumes_by_pool(self):
        pass



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.resources.ds8k.v1.common import types
from ...base import TestCaseWithConnect
from ...test_resources.test_ds8k.base import TestUtils
from pyds8k.resources.ds8k.v1.common.base import Base
from pyds8k.dataParser.ds8k import ResponseParser, RequestParser
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_json_by_type, \
    get_response_data_by_type, \
    get_request_json_body, create_mappings_response_json
from pyds8k.client.ds8k.v1.sc_client import SCClient
import httpretty
from pyds8k.resources.utils import get_resource_class_by_name
from pyds8k.base import Resource

system_list_res_json = get_response_list_json_by_type(types.DS8K_SYSTEM)
system_list_res = get_response_list_data_by_type(types.DS8K_SYSTEM)
volume_list_res_json = get_response_list_json_by_type(types.DS8K_VOLUME)
volume_list_res = get_response_list_data_by_type(types.DS8K_VOLUME)
volume_a_res_json = get_response_json_by_type(types.DS8K_VOLUME)
volume_a_res = get_response_data_by_type(types.DS8K_VOLUME)


class TestClient(TestUtils, TestCaseWithConnect):

    def setUp(self):
        super(TestClient, self).setUp()
        self.base_url = Base.base_url
        self.rest_client = SCClient('localhost', 'admin', 'admin', '8088')

    def _assert_equal_between_dicts(self,
                                    returned_dict,
                                    origin_dict
                                    ):
        for key, value in origin_dict.items():
            if not isinstance(value, dict):
                self.assertEqual(value, returned_dict.get(key))

    def _assert_equal_between_obj_and_dict(self,
                                           returned_obj,
                                           origin_dict
                                           ):
        for key, value in origin_dict.items():
            if not isinstance(value, dict):
                self.assertEqual(value, getattr(returned_obj, key))

    def _set_resource_list(self, route):
        resource_response = get_response_data_by_type(route)
        prefix = '{}.{}'.format(self.client.service_type,
                                self.client.service_version
                                )
        res_class, _ = get_resource_class_by_name(str(route).lower(),
                                                  prefix
                                                  )
        if res_class.__name__ == Resource.__name__:
            raise Exception(
                'Can not get resource class from route: {}'.format(route)
                )
        id_field = res_class.id_field
        route_id = self._get_resource_id_from_resopnse(route,
                                                       resource_response,
                                                       id_field
                                                       )
        url = '/{}/{}'.format(route, route_id)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=get_response_json_by_type(route),
                               content_type='application/json',
                               status=200,
                               )
        return route_id

    def _set_sub_resource(self, route, route_id, sub_route):
        sub_route_url = '/{}/{}/{}'.format(route, route_id, sub_route)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + sub_route_url,
                               body=get_response_list_json_by_type(sub_route),
                               content_type='application/json',
                               status=200,
                               )

    def _post_sub_resource(self, route, route_id, sub_route, body):
        sub_route_url = '/{}/{}/{}'.format(route, route_id, sub_route)

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + sub_route_url)

            resq = RequestParser(body)
            self.assertEqual(get_request_json_body(request.body),
                             resq.get_request_data())
            return (200, headers, create_mappings_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + sub_route_url,
                               body=_verify_request,
                               content_type='application/json',
                               )

    @httpretty.activate
    def _test_resource_by_route(self, route, func, sub_resource=[]):
        route_id = self._set_resource_list(route)
        if sub_resource:
            for sub_route in sub_resource:
                self._set_sub_resource(route, route_id, sub_route)
        res = getattr(self.rest_client, func)(route_id)[0]
        self.assertIsInstance(res, dict)
        rep = ResponseParser(get_response_data_by_type(route),
                             route).get_representations()[0]
        self._assert_equal_between_dicts(res, rep)

    @httpretty.activate
    def _test_sub_resource(self, route, sub_route, func):
        route_id = self._set_resource_list(route)
        self._set_sub_resource(route, route_id, sub_route)
        res = getattr(self.rest_client, func)(route_id)[0]
        self.assertIsInstance(res, dict)
        # print "&&&&&&&&&&&{}".format(res)
        rep = ResponseParser(get_response_data_by_type(sub_route),
                             sub_route).get_representations()[0]
        self._assert_equal_between_dicts(res, rep)

    @httpretty.activate
    def _test_resource_list_by_route(self, route, func=None):
        prefix = '{}.{}'.format(self.client.service_type,
                                self.client.service_version
                                )
        res_class, _ = get_resource_class_by_name(str(route).lower(),
                                                  prefix
                                                  )
        if res_class.__name__ == Resource.__name__:
            raise Exception(
                'Can not get resource class from route: {}'.format(route)
                )
        url = '/{}'.format(route)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=get_response_list_json_by_type(route),
                               content_type='application/json',
                               status=200,
                               )
        func = func or 'get_{}'.format(route)
        res = getattr(self.rest_client, func)()
        self.assertIsInstance(res, list)
        self.assertIsInstance(res[0], dict)
        # print "&&&&&&&&&&&{}".format(res[0])
        rep = ResponseParser(get_response_list_data_by_type(route),
                             route).get_representations()[0]
        self._assert_equal_between_dicts(res[0], rep)

    @httpretty.activate
    def _test_sub_resource_post(self, route, sub_route, func, body, *params):
        route_id = self._set_resource_list(route)
        self._post_sub_resource(route, route_id, sub_route, body)
        func = func or 'get_{}'.format(route)
        res = getattr(self.rest_client, func)(route_id, *params)[0]
        rep = ResponseParser(get_response_data_by_type(sub_route),
                             sub_route).get_representations()[0]
        self._assert_equal_between_obj_and_dict(res, rep)

    @httpretty.activate
    def test_get_system(self):
        sys_url = '/systems'

        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + sys_url,
                               body=system_list_res_json,
                               content_type='application/json')

        sys = self.rest_client.get_system()[0]
        self.assertIsInstance(sys, dict)
        rep = ResponseParser(system_list_res,
                             types.DS8K_SYSTEM).get_representations()[0]
        self._assert_equal_between_dicts(sys, rep)

    def test_get_volume(self):
        self._test_resource_by_route(types.DS8K_VOLUME, 'get_volume')

    def test_get_extentpool(self):
        self._test_resource_by_route(types.DS8K_POOL, 'get_extentpool',
                                     sub_resource=[types.DS8K_ESEREP, ])

    def test_list_extentpools(self):
        self._test_resource_list_by_route(types.DS8K_POOL, 'list_extentpools')

    def test_list_extentpool_volumes(self):
        self._test_sub_resource(types.DS8K_POOL,
                                types.DS8K_VOLUME,
                                'list_extentpool_volumes',
                                )

    def test_list_extentpool_virtualpool(self):
        self._test_sub_resource(types.DS8K_POOL,
                                types.DS8K_ESEREP,
                                'list_extentpool_virtualpool',
                                )

    def test_list_flashcopies(self):
        self._test_resource_list_by_route(types.DS8K_FLASHCOPY,
                                          'list_flashcopies'
                                          )

    def test_list_volume_flashcopies(self):
        self._test_sub_resource(types.DS8K_VOLUME,
                                types.DS8K_FLASHCOPY,
                                'list_volume_flashcopies',
                                )

    def test_list_remotecopies(self):
        self._test_resource_list_by_route(types.DS8K_PPRC,
                                          'list_remotecopies'
                                          )

    def test_list_volume_remotecopies(self):
        self._test_sub_resource(types.DS8K_VOLUME,
                                types.DS8K_PPRC,
                                'list_volume_remotecopies',
                                )

    def test_list_logical_subsystems(self):
        self._test_resource_list_by_route(types.DS8K_LSS,
                                          'list_logical_subsystems'
                                          )

    def test_list_lss_volumes(self):
        self._test_sub_resource(types.DS8K_LSS,
                                types.DS8K_VOLUME,
                                'list_lss_volumes',
                                )

    def test_list_fcports(self):
        self._test_resource_list_by_route(types.DS8K_IOPORT,
                                          'list_fcports'
                                          )

    def test_map_volume_to_host(self):
        volume_id = '000B'
        lunid = '09'
        body = {"mappings": [{lunid: volume_id}, ]}
        self._test_sub_resource_post(types.DS8K_HOST,
                                     types.DS8K_VOLMAP,
                                     'map_volume_to_host',
                                     body,
                                     volume_id,
                                     lunid
                                     )
        body = {"volumes": [volume_id]}
        self._test_sub_resource_post(types.DS8K_HOST,
                                     types.DS8K_VOLMAP,
                                     'map_volume_to_host',
                                     body,
                                     volume_id,
                                     ''
                                     )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
from functools import cmp_to_key
from pyds8k.dataParser.ds8k import RequestParser
from pyds8k.resources.ds8k.v1.common.types import DS8K_POOL, \
    DS8K_VOLUME, \
    DS8K_TSEREP, \
    DS8K_ESEREP
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_json_by_type, \
    get_response_data_by_type
from ...data import action_response, action_response_json
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.volumes import Volume
from pyds8k.resources.ds8k.v1.pools import Pool
from pyds8k.resources.ds8k.v1.tserep import TSERep
from pyds8k.resources.ds8k.v1.eserep import ESERep

response_a = get_response_data_by_type(DS8K_POOL)
response_a_json = get_response_json_by_type(DS8K_POOL)


class TestPool(TestDS8KWithConnect):

    def setUp(self):
        super(TestPool, self).setUp()
        self.pool_id = self._get_resource_id_from_resopnse(DS8K_POOL,
                                                           response_a,
                                                           Pool.id_field
                                                           )
        self.pool = self.system.one(
            DS8K_POOL,
            self.pool_id,
            rebuild_url=True
        )

    def test_get_volumes(self):
        self._test_sub_resource_list_by_route(DS8K_POOL, DS8K_VOLUME,
                                              self._sorted_by_volume_name
                                              )

    def test_get_tserep(self):
        self._test_sub_resource_list_by_route(DS8K_POOL, DS8K_TSEREP)

    def test_get_eserep(self):
        self._test_sub_resource_list_by_route(DS8K_POOL, DS8K_ESEREP)

    @httpretty.activate
    def test_delete_tserep(self):
        url = '/pools/{}/tserep'.format(self.pool_id)
        httpretty.register_uri(
            httpretty.DELETE,
            self.domain + self.base_url + url,
            content_type='application/json',
            status=204,
        )
        self.pool.delete_tserep()
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_delete_eserep(self):
        url = '/pools/{}/eserep'.format(self.pool_id)
        httpretty.register_uri(
            httpretty.DELETE,
            self.domain + self.base_url + url,
            content_type='application/json',
            status=204,
        )
        self.pool.delete_eserep()
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_update_tserep_cap(self):
        url = '/pools/{}/tserep'.format(self.pool_id)
        cap = '10'
        captype = 'gib'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'cap': cap, 'captype': captype})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        _, body = self.pool.update_tserep_cap(cap, captype)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(body, action_response['server'])

    @httpretty.activate
    def test_update_tserep_threshold(self):
        url = '/pools/{}/tserep'.format(self.pool_id)
        threshold = '70'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'threshold': threshold})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, action_response_json)

        httpretty.register_uri(httpretty.PUT,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        _, body = self.pool.update_tserep_threshold(threshold)
        self.assertEqual(httpretty.PUT, httpretty.last_request().method)
        self.assertEqual(body, action_response['server'])

    def test_update_eserep_cap(self):
        pass

    def test_update_eserep_threshold(self):
        pass

    def test_set_related_resources_collection(self):
        volumes = [Volume(self.client, resource_id='volume{}'.format(i))
                   for i in range(10)
                   ]
        tserep = [TSERep(self.client, info={'pool': {'name': 'testpool_0'}}), ]
        eserep = [ESERep(self.client, info={'pool': {'name': 'testpool_0'}}), ]

        # init without related_resources collection
        pool = Pool(self.client, info={
            'name': 'testpool_0',
            'eserep': '',
            'tserep': '',
            'volumes': {
                'link': {
                    'rel': 'self',
                    'href': '/api/volumes'
                },
            }
        }
                    )
        for i in pool.related_resources_collection:
            self.assertEqual('', pool.representation.get(i))
            self.assertFalse(hasattr(pool, i))

        # loading related resources collection
        pool._start_updating()
        for item in ((DS8K_VOLUME, volumes),
                     (DS8K_TSEREP, tserep),
                     (DS8K_ESEREP, eserep),
                     ):
            setattr(pool, item[0], item[1])
            for j, value in enumerate(pool.representation[item[0]]):
                self.assertEqual(value,
                                 getattr(item[1][j], item[1][j].id_field)
                                 )
        pool._stop_updating()

    @httpretty.activate
    def test_lazy_loading_related_resources_collection(self):
        url = '/pools/{}'.format(self.pool_id)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=response_a_json,
                               content_type='application/json',
                               status=200,
                               )
        for item in Pool.related_resources_collection:
            sub_route_url = '{}/{}'.format(url, item)
            httpretty.register_uri(httpretty.GET,
                                   self.domain + self.base_url + sub_route_url,
                                   body=get_response_list_json_by_type(item),
                                   content_type='application/json',
                                   status=200,
                                   )
        pool = self.system.get_pool(self.pool_id)

        for item in Pool.related_resources_collection:
            res_collection = getattr(pool, item)
            self.assertNotEqual(0, len(res_collection))
            res_collection.sort(
                key=cmp_to_key(
                    self._get_sort_func_by(res_collection[0].id_field))
            )
            res_collection_data = list(
                get_response_list_data_by_type(item)['data'][item]
            )
            res_collection_data.sort(
                key=cmp_to_key(
                    self._get_sort_func_by(res_collection[0].id_field))
            )

            self.assertEqual(
                len(res_collection_data),
                len(res_collection))
            self._assert_equal_between_sorted_dict_and_resource_list(
                res_collection_data,
                res_collection
            )

    def test_set_related_resources_collection_during_loading(self):
        pool = Pool(self.client, info={
            'name': 'testpool_0',
            'volumes': [{
                'id': '0000',
                'link': {
                    'rel': 'self',
                    'href': '/api/volumes/0000'
                },
            },
            ],
        }
                    )

        self.assertEqual('0000', pool.representation.get('volumes')[0])
        self.assertEqual('0000', pool.volumes[0].id)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
    "server": {
        "status": "ok",
        "code": "CMUC00183I",
        "message": "Operation done successfully."
    },
    "counts": {
        "data_counts": 32,
        "total_counts": 32
    },
    'data': {
        'hosts': [
            {
                'name': 'host1',
                'state': 'Online',
                'hosttype': 'VMWare',
                'addrmode': 'SCSI map',
                'addrdiscovery': 'Report LUNs',
                'lbs': '512',
                'host_ports': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/host_ports',
                    },
                },
                'ioports': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/ioports',
                    },
                },
                'volumes': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes',
                    },
                },
                'link': {
                    'rel': 'self',
                    'href': 'https://localhost:8088/api/v1/hosts/host1'
                },
            },
            {
                'name': 'host2',
                'state': 'Online',
                'hosttype': 'VMware',
                'addrmode': 'SCSI map',
                'addrdiscovery': 'Report LUNs',
                'lbs': '512',
                'host_ports': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/host_ports',
                    },
                },
                'ioports': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/ioports',
                    },
                },
                'volumes': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes',
                    },
                },
                'link': {
                    'rel': 'self',
                    'href': 'https://localhost:8088/api/v1/hosts/host2'
                },
            },
        ]
    }
}

ONE = {
    "server": {
        "status": "ok",
        "code": "CMUC00183I",
        "message": "Operation done successfully."
    },
    "counts": {
        "data_counts": 1,
        "total_counts": 1
    },
    'data': {
        'hosts': [
            {
                'name': 'host1',
                'state': 'Online',
                'hosttype': 'VMWare',
                'addrmode': 'SCSI map',
                'addrdiscovery': 'Report LUNs',
                'lbs': '512',
                'host_ports': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/host_ports',
                    },
                },
                'ioports': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/ioports',
                    },
                },
                'volumes': {
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes',
                    },
                },
                'link': {
                    'rel': 'self',
                    'href': 'https://localhost:8088/api/v1/hosts/host1'
                },
            },
        ]
    }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
from pyds8k.exceptions import InternalServerError
from pyds8k.resources.ds8k.v1.common.types import DS8K_HOST, \
    DS8K_VOLMAP
from ...data import get_response_json_by_type, \
    get_response_data_by_type
from ...data import action_response_json, \
    create_mappings_response_json, create_mapping_response_json, \
    action_response_failed, action_response_failed_json
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.mappings import Volmap
from pyds8k.resources.ds8k.v1.hosts import Host
from pyds8k.resources.ds8k.v1.volumes import Volume
from pyds8k.dataParser.ds8k import RequestParser

host_response = get_response_data_by_type(DS8K_HOST)
host_response_json = get_response_json_by_type(DS8K_HOST)
mapping_response = get_response_data_by_type(DS8K_VOLMAP)
mapping_response_json = get_response_json_by_type(DS8K_VOLMAP)


class TestVolmap(TestDS8KWithConnect):

    def setUp(self):
        super(TestVolmap, self).setUp()
        self.host_id = self._get_resource_id_from_resopnse(DS8K_HOST,
                                                           host_response,
                                                           Host.id_field
                                                           )
        self.lunid = self._get_resource_id_from_resopnse(DS8K_VOLMAP,
                                                         mapping_response,
                                                         Volmap.id_field
                                                         )
        self.host = self.system.one(
            DS8K_HOST,
            self.host_id,
            rebuild_url=True,
        )

    def test_related_resource_field(self):
        mapping_info = mapping_response['data'][DS8K_VOLMAP][0]
        volume_id = mapping_info['volume'][Volume.id_field]
        mapping = Volmap(self.client, info=mapping_info)
        self.assertEqual(mapping.volume, volume_id)
        self.assertEqual(mapping.representation['volume'], volume_id)
        self.assertIsInstance(mapping._volume, Volume)
        self.assertEqual(mapping._volume.id, volume_id)

    def test_get_mappings(self):
        self._test_sub_resource_list_by_route(
            DS8K_HOST, DS8K_VOLMAP,
            self._get_sort_func_by(Volmap.id_field)
        )

    @httpretty.activate
    def test_delete_mapping(self):
        url = '/hosts/{}/mappings/{}'.format(self.host_id, self.lunid)
        httpretty.register_uri(
            httpretty.GET,
            self.domain + self.base_url + url,
            body=mapping_response_json,
            content_type='application/json',
            status=200,
        )
        httpretty.register_uri(
            httpretty.DELETE,
            self.domain + self.base_url + url,
            body=action_response_json,
            content_type='application/json',
            status=204,
        )
        # Way 1
        _ = self.host.delete_mapping(self.lunid)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)
        # self.assertEqual(resp1, action_response['server'])

        # Way 2
        mapping = self.host.get_mapping(self.lunid)
        self.assertIsInstance(mapping, Volmap)
        resp2, _ = mapping.delete()
        self.assertEqual(resp2.status_code, 204)
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_delete_mapping_failed(self):
        url = '/hosts/{}/mappings/{}'.format(self.host_id, self.lunid)
        httpretty.register_uri(
            httpretty.DELETE,
            self.domain + self.base_url + url,
            body=action_response_failed_json,
            content_type='application/json',
            status=500,
        )
        with self.assertRaises(InternalServerError) as cm:
            self.host.delete_mapping(self.lunid)
        self.assertEqual(action_response_failed['server'],
                         cm.exception.error_data
                         )
        self.assertEqual(httpretty.DELETE, httpretty.last_request().method)

    @httpretty.activate
    def test_create_mappings_with_volume_id(self):
        url = '/hosts/{}/mappings'.format(self.host_id)
        volumes = ['000{}'.format(i) for i in range(10)]

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'volumes': volumes})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, create_mappings_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.host.create_mappings(volumes=volumes)
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp1[0], Volmap)

    @httpretty.activate
    def test_create_mappings_with_mappings(self):
        url = '/hosts/{}/mappings'.format(self.host_id)
        mappings = [{'0{}'.format(i): '000{}'.format(i)} for i in range(10)]

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'mappings': mappings})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, create_mappings_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        # Way 1
        resp1 = self.host.create_mappings(mappings=mappings)
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(resp1[0], Volmap)

    @httpretty.activate
    def test_create_mapping_with_volume_and_lunid(self):
        url = '/hosts/{}/mappings'.format(self.host_id)
        lunid = '00'
        volume_id = '0000'

        def _verify_request(request, uri, headers):
            self.assertEqual(uri, self.domain + self.base_url + url)

            resq = RequestParser({'lunid': lunid, 'volume': volume_id})
            self.assertEqual(json.loads(request.body), resq.get_request_data())
            return (200, headers, create_mapping_response_json)

        httpretty.register_uri(httpretty.POST,
                               self.domain + self.base_url + url,
                               body=_verify_request,
                               content_type='application/json',
                               )
        mapping = self.host.all(DS8K_VOLMAP)
        new_mapping = mapping.create(lunid=lunid, volume=volume_id)
        resp, data = new_mapping.save()
        self.assertEqual(httpretty.POST, httpretty.last_request().method)
        self.assertIsInstance(data[0], Volmap)
        self.assertEqual(resp.status_code, 200)

        @httpretty.activate
        def test_create_mapping_with_volume(self):
            url = '/hosts/{}/mappings'.format(self.host_id)
            volume_id = '0000'

            def _verify_request(request, uri, headers):
                self.assertEqual(uri, self.domain + self.base_url + url)

                resq = RequestParser({'lunid': '', 'volume': volume_id})
                self.assertEqual(json.loads(request.body),
                                 resq.get_request_data())
                return (200, headers, create_mapping_response_json)

            httpretty.register_uri(httpretty.POST,
                                   self.domain + self.base_url + url,
                                   body=_verify_request,
                                   content_type='application/json',
                                   )
            mapping = self.host.all(DS8K_VOLMAP)
            new_mapping = mapping.create(lunid='', volume=volume_id)
            resp, data = new_mapping.save()
            self.assertEqual(httpretty.POST, httpretty.last_request().method)
            self.assertIsInstance(data[0], Volmap)
            self.assertEqual(resp.status_code, 200)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from collections import namedtuple

Device = namedtuple('Device', ['ipaddr', 'username', 'password', 'port'])

ds44 = Device(ipaddr='9.11.108.44',
              username='admin',
              password='open1sys',
              port=8452
              )

ds179 = Device(ipaddr='9.11.217.179',
               username='admin',
               password='passw0rd',
               port=8452
               )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

DS8K_TOKEN = 'tokens'
DS8K_SYSTEM = 'systems'
DS8K_NODE = 'nodes'
DS8K_MARRAY = 'marrays'
DS8K_USER = 'users'
DS8K_ENCRYPTION_GROUP = 'encryption_groups'
DS8K_IOENCLOSURE = 'io_enclosures'
DS8K_POOL = 'pools'
DS8K_TSEREP = 'tserep'
DS8K_ESEREP = 'eserep'
DS8K_LSS = 'lss'
DS8K_VOLUME = 'volumes'
DS8K_SE = 'SE'
DS8K_VOLUME_GROUP = 'volgrps'
DS8K_HOST_PORT = 'host_ports'
DS8K_HOST = 'hosts'
DS8K_IOPORT = 'ioports'
DS8K_VOLMAP = 'mappings'
DS8K_EVENT = 'events'
DS8K_VOLUME_TYPE_FB = 'fb'
DS8K_VOLUME_TYPE_CKD = 'ckd'
DS8K_VOLUME_TYPES = (DS8K_VOLUME_TYPE_FB, DS8K_VOLUME_TYPE_CKD)
DS8K_FLASHCOPY = 'flashcopy'
DS8K_PPRC = 'pprc'
DS8K_CAPTYPE_GIB = 'gib'
DS8K_CAPTYPE_BYTE = 'bytes'
DS8K_CAPTYPE_CYL = 'cyl'
DS8K_CAPTYPE_MOD1 = 'mod1'
DS8K_CAPTYPES = (DS8K_CAPTYPE_GIB,
                 DS8K_CAPTYPE_BYTE,
                 DS8K_CAPTYPE_CYL,
                 DS8K_CAPTYPE_MOD1
                 )
DS8K_TP_NONE = 'none'
DS8K_TP_ESE = 'ese'
DS8K_TP_TSE = 'tse'
DE8K_TPS = (DS8K_TP_NONE, DS8K_TP_ESE, DS8K_TP_TSE)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
       "server":
       {
           "status": "ok",
           "code": "CMUC00183I",
           "message": "Operation done successfully."
       },
       "counts":
       {
           "data_counts": 5,
           "total_counts": 5
       },
       "data":
       {
           "lss":
           [
               {
                   "id": "00",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/lss/00"
                   },
                   "group": "0",
                   "addrgrp": "0",
                   "type": "fb",
                   "configvols": "25",
                   "volumes":
                   {
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/"
                                   "lss/00/volumes"
                       }
                   }
               },
               {
                   "id": "02",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/"
                               "lss/02"
                   },
                   "group": "0",
                   "addrgrp": "0",
                   "type": "fb",
                   "configvols": "16",
                   "volumes":
                   {
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/"
                                   "lss/02/volumes"
                       }
                   }
               },
               {
                   "id": "04",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/lss/04"
                   },
                   "group": "0",
                   "addrgrp": "0",
                   "type": "fb",
                   "configvols": "256",
                   "volumes":
                   {
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/"
                                   "lss/04/volumes"
                       }
                   }
               },
               {
                   "id": "06",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/lss/06"
                   },
                   "group": "0",
                   "addrgrp": "0",
                   "type": "fb",
                   "configvols": "256",
                   "volumes":
                   {
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/"
                                   "lss/06/volumes"
                       }
                   }
               },
               {
                   "id": "08",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/lss/08"
                   },
                   "group": "0",
                   "addrgrp": "0",
                   "type": "fb",
                   "configvols": "256",
                   "volumes":
                   {
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/"
                                   "lss/08/volumes"
                       }
                   }
               }
           ]
       }
    }


ONE = {
       "server":
       {
           "status": "ok",
           "code": "CMUC00183I",
           "message": "Operation done successfully."
       },
       "counts":
       {
           "data_counts": 1,
           "total_counts": 1
       },
       "data":
       {
           "lss":
           [
               {
                   "id": "00",
                   "link":
                   {
                       "rel": "self",
                       "href": "https://localhost:8088/api/v1/lss/00"
                   },
                   "group": "0",
                   "addrgrp": "0",
                   "type": "fb",
                   "configvols": "25",
                   "volumes":
                   {
                       "link":
                       {
                           "rel": "self",
                           "href": "https://localhost:8088/api/v1/lss/"
                                   "00/volumes"
                       }
                   }
               }
           ]
       }
    }



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Host Ports interface.
"""
from .common.types import DS8K_HOST_PORT
from .common.base import Base, BaseManager
from .hosts import Host, HostManager
from .ioports import IOPort, IOPortManager


class HostPort(Base):
    id_field = 'wwpn'
    _template = {'wwpn': '',
                 'state': None,
                 'hosttype': None,
                 'addrdiscovery': None,
                 'lbs': None,
                 'host': '',
                 'login_type': None,
                 'logical_path_established': None,
                 'wwnn': None,
                 'login_ports': None,
                 }
    related_resource = {'_host': (Host, HostManager),
                        }
    readonly_fileds = ('state', 'hosttype', 'addrdiscovery', 'lbs',)

    def _add_details(self, info, force=False):
        super(HostPort, self)._add_details(info, force=force)
        self._start_updating()
        self._set_ioports()
        self._stop_updating()

    def _set_ioports(self):
        OCCUPIED_IOPORTS = 'login_ports'
        port_list = self.representation.get(OCCUPIED_IOPORTS, [])
        port_obj_list = []
        for port in port_list:
            port_obj = IOPort(self.client,
                              manager=IOPortManager(self.client),
                              info=port,
                              loaded=False,
                              )
            port_obj_list.append(port_obj)
        self.representation[OCCUPIED_IOPORTS] = [p.id for p in port_obj_list]
        setattr(self, OCCUPIED_IOPORTS, port_obj_list)

    # def __repr__(self):
    #    return "<HostPort: {0}>".format(self.id)


class HostPortManager(BaseManager):
    """
    Manage Host Ports resources.
    """
    resource_class = HostPort
    resource_type = DS8K_HOST_PORT

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id,
                         url=url, obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)


RESOURCE_TUPLE = (HostPort, HostPortManager)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
from functools import cmp_to_key
from pyds8k.resources.ds8k.v1.common.types import DS8K_LSS, \
    DS8K_VOLUME
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_json_by_type
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.lss import LSS, LSSManager
from pyds8k.resources.ds8k.v1.volumes import Volume


class TestLSS(TestDS8KWithConnect):

    def setUp(self):
        super(TestLSS, self).setUp()
        self.lss = LSS(self.client, LSSManager(self.client))

    def test_get_volumes(self):
        self._test_sub_resource_list_by_route(DS8K_LSS, DS8K_VOLUME,
                                              self._sorted_by_volume_name
                                              )

    def test_set_related_resources_collection(self):
        volumes = [Volume(self.client, resource_id='volume{}'.format(i))
                   for i in range(10)
                   ]

        # init without related_resources collection
        lss = LSS(self.client, info={
            'volumes': {
                'link': {
                    'rel': 'self',
                    'href': '/api/volumes'
                },
            }
        }
                  )
        for i in lss.related_resources_collection:
            self.assertEqual('', lss.representation.get(i))
            self.assertFalse(hasattr(lss, i))

        # loading related resources collection
        lss._start_updating()
        for item in ((DS8K_VOLUME, volumes),
                     ):
            setattr(lss, item[0], item[1])
            for j, value in enumerate(lss.representation[item[0]]):
                self.assertEqual(value,
                                 getattr(item[1][j], item[1][j].id_field)
                                 )
        lss._stop_updating()

    @httpretty.activate
    def test_lazy_loading_related_resources_collection(self):
        lss_id = '00'
        url = '/lss/{}'.format(lss_id)
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=get_response_json_by_type(DS8K_LSS),
                               content_type='application/json',
                               status=200,
                               )
        for item in LSS.related_resources_collection:
            sub_route_url = '{}/{}'.format(url, item)
            httpretty.register_uri(httpretty.GET,
                                   self.domain + self.base_url + sub_route_url,
                                   body=get_response_list_json_by_type(item),
                                   content_type='application/json',
                                   status=200,
                                   )
        lss = self.system.get_lss_by_id(lss_id)

        for item in LSS.related_resources_collection:
            res_collection = getattr(lss, item)
            self.assertNotEqual(0, len(res_collection))
            res_collection.sort(
                key=cmp_to_key(
                    self._get_sort_func_by(res_collection[0].id_field))
            )
            res_collection_data = list(
                get_response_list_data_by_type(item)['data'][item]
            )
            res_collection_data.sort(
                key=cmp_to_key(
                    self._get_sort_func_by(res_collection[0].id_field))
            )

            self.assertEqual(
                len(res_collection_data),
                len(res_collection))
            self._assert_equal_between_sorted_dict_and_resource_list(
                res_collection_data,
                res_collection
            )

    def test_set_related_resources_collection_during_loading(self):
        lss = LSS(self.client, info={
            'volumes': [{
                'id': '0000',
                'link': {
                    'rel': 'self',
                    'href': '/api/volumes/0000'
                },
            },
            ],
        }
                  )

        self.assertEqual('0000', lss.representation.get('volumes')[0])
        self.assertEqual('0000', lss.volumes[0].id)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 6,
    "total_counts": 6
  },
  "data": {
    "pools": [
      {
        "id": "P1",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P1"
        },
        "name": "PoolName_ckd1_b",
        "node": "1",
        "stgtype": "ckd",
        "cap": "439635",
        "capalloc": "25599",
        "capavail": "414036",
        "overprovisioned": "",
        'real_capacity_allocated_on_ese': '0',
        'virtual_capacity_allocated_on_ese': '0',
        "easytier": "none",
        "tieralloc":
               [
                   {
                       "tier": "ENT",
                       "assigned": "0",
                       "cap": "446676598784",
                       "allocated": "0"
                   }
               ],
        "threshold": "15",
        "eserep": {
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P1/eserep"
          }
        },
        "tserep": {
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P1/tserep"
          }
        },
        "volumes": ""
      },
      {
        "id": "P2",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P2"
        },
        "name": "regression",
        "node": "0",
        "stgtype": "fb",
        "cap": "512",
        "capalloc": "512",
        "capavail": "512",
        "overprovisioned": "0.6",
        'real_capacity_allocated_on_ese': '0',
        'virtual_capacity_allocated_on_ese': '0',
        "threshold": "15",
        "eserep": "",
        "tserep": {
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2/tserep"
          }
        },
        "volumes": {
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P2/volumes"
          }
        }
      },
      {
        "id": "P3",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P3"
        },
        "name": "vPool",
        "node": "1",
        "stgtype": "ckd",
        "cap": "518658",
        "capalloc": "0",
        "capavail": "518658",
        "overprovisioned": "0.0",
        'real_capacity_allocated_on_ese': '0',
        'virtual_capacity_allocated_on_ese': '0',
        "threshold": "15",
        "eserep": "",
        "tserep": "",
        "volumes": ""
      },
      {
        "id": "P4",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P4"
        },
        "name": "p5",
        "node": "0",
        "stgtype": "fb",
        "cap": "0",
        "capalloc": "0",
        "capavail": "0",
        "overprovisioned": "",
        'real_capacity_allocated_on_ese': '0',
        'virtual_capacity_allocated_on_ese': '0',
        "easytier": "none",
        "tieralloc":
               [
                   {
                       "tier": "ENT",
                       "assigned": "0",
                       "cap": "446676598784",
                       "allocated": "0"
                   }
               ],
        "threshold": "15",
        "eserep": "",
        "tserep": "",
        "volumes": ""
      },
      {
        "id": "P6",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P6"
        },
        "name": "fb_odd",
        "node": "0",
        "stgtype": "fb",
        "cap": "0",
        "capalloc": "0",
        "capavail": "0",
        "overprovisioned": "",
        'real_capacity_allocated_on_ese': '0',
        'virtual_capacity_allocated_on_ese': '0',
        "easytier": "none",
        "tieralloc":
               [
                   {
                       "tier": "ENT",
                       "assigned": "0",
                       "cap": "446676598784",
                       "allocated": "0"
                   }
               ],
        "threshold": "15",
        "eserep": "",
        "tserep": "",
        "volumes": ""
      }
    ]
  }
}

ONE = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "counts": {
    "data_counts": 1,
    "total_counts": 1
  },
  "data": {
    "pools": [
      {
        "id": "P1",
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P1"
        },
        "name": "PoolName_ckd1_b",
        "node": "1",
        "stgtype": "ckd",
        "cap": "439635",
        "capalloc": "25599",
        "capavail": "414036",
        "overprovisioned": "",
        'real_capacity_allocated_on_ese': '0',
        'virtual_capacity_allocated_on_ese': '0',
        "easytier": "none",
        "tieralloc":
               [
                   {
                       "tier": "ENT",
                       "assigned": "0",
                       "cap": "446676598784",
                       "allocated": "0"
                   }
               ],
        "threshold": "15",
        "eserep": {
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P1/eserep"
          }
        },
        "tserep": {
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P1/tserep"
          }
        },
        "volumes": ""
      },
    ]
  }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
    "server": {
        "status": "ok",
        "code": "CMUC00183I",
        "message": "Operation done successfully."
    },
    "counts": {
        "data_counts": 32,
        "total_counts": 32
    },
    'data': {
        'flashcopy': [
            {
                'sourcevolume': {
                    'id': '0000',
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes/0000',
                    },
                },
                'targetvolume': {
                    'id': '0001',
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes/0001',
                    },
                },
                'persistent': 'enabled',
                'recording': 'enabled',
                'backgroundcopy': 'enabled',
                'state': 'valid',
            },
            {
                'sourcevolume': {
                    'id': '1000',
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes/1000',
                    },
                },
                'targetvolume': {
                    'id': '1001',
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes/1001',
                    },
                },
                'persistent': 'enabled',
                'recording': 'enabled',
                'backgroundcopy': 'enabled',
                'state': 'valid',
            },
        ]
    }
}

ONE = {
    "server": {
        "status": "ok",
        "code": "CMUC00183I",
        "message": "Operation done successfully."
    },
    "counts": {
        "data_counts": 1,
        "total_counts": 1
    },
    'data': {
        'flashcopy': [
            {
                'sourcevolume': {
                    'id': '0000',
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes/0000',
                    },
                },
                'targetvolume': {
                    'id': '0001',
                    'link': {
                        'rel': 'self',
                        'href': 'https://localhost:8088/api/v1/volumes/0001',
                    },
                },
                'persistent': 'enabled',
                'recording': 'enabled',
                'backgroundcopy': 'enabled',
                'state': 'valid',
            },
        ]
    }
}



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
from pyds8k.resources.ds8k.v1.common.types import DS8K_SYSTEM
from .base import TestDS8KWithConnect
from pyds8k.resources.ds8k.v1.systems import System, \
    SystemManager
from ...data import get_response_list_json_by_type, \
    get_response_list_data_by_type
from pyds8k.exceptions import OperationNotAllowed

system_list_response = get_response_list_data_by_type(DS8K_SYSTEM)
system_list_response_json = get_response_list_json_by_type(DS8K_SYSTEM)


class TestSystem(TestDS8KWithConnect):

    def setUp(self):
        super(TestSystem, self).setUp()
        self.system = System(self.client, SystemManager(self.client))

    @httpretty.activate
    def test_get_system(self):
        url = '/systems'
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=system_list_response_json,
                               content_type='application/json',
                               status=200,
                               )
        sys = self.system.get_system()
        self.assertIsInstance(sys, System)
        sys_data = system_list_response['data']['systems'][0]
        self._assert_equal_between_dict_and_resource(sys_data, sys)

    @httpretty.activate
    def test_not_allowed_operations(self):
        url = '/systems'
        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=system_list_response_json,
                               content_type='application/json',
                               status=200,
                               )
        sys = self.system.get_system()
        with self.assertRaises(OperationNotAllowed):
            sys.put()
        with self.assertRaises(OperationNotAllowed):
            sys.patch()
        with self.assertRaises(OperationNotAllowed):
            sys.posta()
        with self.assertRaises(OperationNotAllowed):
            sys.delete()
        with self.assertRaises(OperationNotAllowed):
            sys.update()
        with self.assertRaises(OperationNotAllowed) as cm:
            sys.save()
        self.assertEqual(System.__name__, cm.exception.resource_name)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

GiB = 2**30     # Gibibyte = GiB = 2^30 B = 1,073,741,824 bytes
MiB = 2**20     # Mebibyte = MiB = 2^20 B = 1,048,576 bytes
KiB = 2**10     # Kibibyte = kiB = 2^10 B = 1,024 bytes

GB = 10**9      # Gigabyte = GB = 10^9 B = 1,000,000,000 bytes
MB = 10**6      # Megabyte = MB = 10^6 B = 1,000,000 bytes
KB = 10**3      # Kilobyte = kB = 10^3 B = 1,000 bytes


def validate_number(number):
    if not isinstance(number, (int, float)):
        raise ValueError("Expected types are (int, long, float)")

# =============================================================================
# Methods converting to bytes.
# =============================================================================


def convert_size_gib_to_bytes(size_in_gib):
    """:rtype: int / long"""
    validate_number(size_in_gib)
    return size_in_gib * GiB


def convert_size_mib_to_bytes(size_in_mib):
    """:rtype: int / long"""
    validate_number(size_in_mib)
    return size_in_mib * MiB


def convert_size_kib_to_bytes(size_in_kib):
    """:rtype: int / long"""
    validate_number(size_in_kib)
    return size_in_kib * KiB


def convert_size_gb_to_bytes(size_in_gb):
    """:rtype: int / long"""
    validate_number(size_in_gb)
    return size_in_gb * GB


def convert_size_mb_to_bytes(size_in_mb):
    """:rtype: int / long"""
    validate_number(size_in_mb)
    return size_in_mb * MB


def convert_size_kb_to_bytes(size_in_kb):
    """:rtype: int / long"""
    validate_number(size_in_kb)
    return size_in_kb * KB


# =============================================================================
# Methods converting from bytes.
# =============================================================================


def convert_size_bytes_to_gib(size_in_bytes):
    """:rtype: float"""
    return float(size_in_bytes) / GiB


def convert_size_bytes_to_mib(size_in_bytes):
    """:rtype: float"""
    return float(size_in_bytes) / MiB


def convert_size_bytes_to_kib(size_in_bytes):
    """:rtype: float"""
    return float(size_in_bytes) / KiB


def convert_size_bytes_to_gb(size_in_bytes):
    """:rtype: float"""
    return float(size_in_bytes) / GB


def convert_size_bytes_to_mb(size_in_bytes):
    """:rtype: float"""
    return float(size_in_bytes) / MB


def convert_size_bytes_to_kb(size_in_bytes):
    """:rtype: float"""
    return float(size_in_bytes) / KB



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
PPRC interface.
"""

from .common.base import Base, ReadOnlyManager
from .common.types import DS8K_PPRC
from .volumes import Volume, VolumeManager
from .systems import System, SystemManager


class PPRC(Base):
    # id_field = 'id'
    _template = {'id': '',
                 'type': '',
                 'state': '',
                 'targetsystem': '',
                 'sourcevolume': '',
                 'targetvolume': '',
                 }

    related_resource = {'_sourcevolume': (Volume, VolumeManager),
                        '_targetvolume': (Volume, VolumeManager),
                        '_targetsystem': (System, SystemManager),
                        }

    def _add_details(self, info, force=False):
        super(PPRC, self)._add_details(info, force=force)

        # Temporarily, remove this line when flashcopy resource has id field.
        self._id = self.representation['id'] = '{}:{}'.format(
            info['sourcevolume']['id'],
            info['targetvolume']['id']
        )

    # def __repr__(self):
    #    return "<FlashCopy: {}>".format(self.id)


class PPRCManager(ReadOnlyManager):
    """
    Manage PPRC resources.
    """
    resource_class = PPRC
    resource_type = DS8K_PPRC


RESOURCE_TUPLE = (PPRC, PPRCManager)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
  "server": {
    "status": "ok",
    "code": "CMUC00183I",
    "message": "Operation done successfully."
  },
  "data": {
    "tserep": [
      {
        "link": {
          "rel": "self",
          "href": "https://localhost:8088/api/v1/pools/P1/tserep"
        },
        "cap": "22260",
        "capalloc": "0",
        "capavail": "22260",
        "overprovisioned": "0.0",
        "repcapthreshold": "80",
        "pool": {
          "id": "P1",
          "link": {
            "rel": "self",
            "href": "https://localhost:8088/api/v1/pools/P1"
          }
        }
      }
    ]
  }
}

ONE = ALL



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
import json
from . import base
from pyds8k import exceptions
from .data import get_response_data_by_type
from pyds8k.resources.ds8k.v1.common.base import Base

default_a_response = get_response_data_by_type('default')
response_401 = {
  "server": {
    "status": "failed",
    "code": "BE742607",
    "message": "The token is invalid or expired."
  }
}

response_token = {
  "server": {
    "status": "ok",
    "code": "",
    "message": "Operation done successfully."
  },
  "token": {
    "token": "54546d2a",
    "expired_time": "2014-08-29T20:13:24+0800",
    "max_idle_interval": "1800000"
  }
}

response_token_error = {
  "server": {
    "status": "failed",
    "code": "NIServerException",
    "message": "Operation done successfully."
  }
}

DEFAULT = 'default'


class TestHTTPException(base.TestCaseWithConnect):

    def setUp(self):
        super(TestHTTPException, self).setUp()
        self.base_url = ''

    @httpretty.activate
    def test_response_status_400(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=400
        )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.BadRequest, vol.get)

    @httpretty.activate
    def test_response_status_401(self):
        domain = self.client.domain
        url = '/default/a'
        _base_url = Base.base_url
        httpretty.register_uri(
            httpretty.POST,
            domain + _base_url + '/tokens',
            body=json.dumps(response_token),
            content_type='application/json',
            status=200
        )

        httpretty.register_uri(
            httpretty.GET,
            domain + self.base_url + url,
            responses=[
                httpretty.Response(body=json.dumps(response_401),
                                   content_type='application/json',
                                   status=401
                                   ),
                httpretty.Response(body=json.dumps(default_a_response),
                                   content_type='application/json',
                                   status=200
                                   ),
                httpretty.Response(body=json.dumps(response_401),
                                   content_type='application/json',
                                   status=401
                                   ),
            ]
        )
        vol = self.resource.one(DEFAULT, 'a')
        vol.get()
        self.assertEqual(
                    vol.url,
                    default_a_response['data']['default'][0]['link']['href']
                         )
        self.assertEqual(
                         vol.name,
                         default_a_response['data']['default'][0]['name']
                         )
        self.assertRaises(exceptions.Unauthorized, vol.get)

    @httpretty.activate
    def test_auth_fail(self):
        domain = self.client.domain
        url = '/default/a'
        _base_url = Base.base_url
        httpretty.register_uri(httpretty.POST,
                               domain + _base_url + '/tokens',
                               body=json.dumps(response_token_error),
                               content_type='application/json',
                               status=401)

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            responses=[
                httpretty.Response(body=json.dumps(response_401),
                                   content_type='application/json',
                                   status=401
                                   ),
                httpretty.Response(body=json.dumps(default_a_response),
                                   content_type='application/json',
                                   status=200
                                   ),
                httpretty.Response(body=json.dumps(response_401),
                                   content_type='application/json',
                                   status=401
                                   ),
            ]
        )
        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.Unauthorized, vol.get)

    @httpretty.activate
    def test_response_status_403(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET,
            domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=403
            )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.Forbidden, vol.get)

    @httpretty.activate
    def test_response_status_404(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=404
        )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.NotFound, vol.get)

    @httpretty.activate
    def test_response_status_405(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=405
            )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.MethodNotAllowed, vol.get)

    @httpretty.activate
    def test_response_status_409(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=409
            )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.Conflict, vol.get)

    @httpretty.activate
    def test_response_status_415(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=415
            )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.UnsupportedMediaType, vol.get)

    @httpretty.activate
    def test_response_status_500(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=500
        )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.InternalServerError, vol.get)

    @httpretty.activate
    def test_response_status_503(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=503
            )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.ServiceUnavailable, vol.get)

    @httpretty.activate
    def test_response_status_504(self):
        domain = self.client.domain
        url = '/default/a'

        httpretty.register_uri(
            httpretty.GET, domain + self.base_url + url,
            body=json.dumps(
                {'server': {'message': 'error', 'details': 'error'}}
            ),
            content_type='application/json',
            status=504
            )

        vol = self.resource.one(DEFAULT, 'a')
        self.assertRaises(exceptions.GatewayTimeout, vol.get)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

ALL = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
        },
    "counts": {
        "data_counts": 2,
        "total_counts": 2
        },
    "data": {
        "encryption_groups": [
            {
                "id": "1",
                "link": {
                    "rel": "self",
                    "href": "https://localhost:8088/api/v1/encryption_groups/1"
                    },
                "state": "accessible",
            },
            {
                "id": "2",
                "link": {
                    "rel": "self",
                    "href": "https://localhost:8088/api/v1/encryption_groups/2"
                    },
                "state": "accessible",
            },
        ]
    }
}


ONE = {
    "server": {
        "status": "ok",
        "code": "",
        "message": "Operation done successfully."
        },
    "counts": {
        "data_counts": 1,
        "total_counts": 1
        },
    "data": {
        "encryption_groups": [
            {
                "id": "1",
                "link": {
                    "rel": "self",
                    "href": "https://localhost:8088/api/v1/encryption_groups/1"
                    },
                "state": "accessible",
            },
        ]
    }
}
